
/*
   Tile Studio Scroll Demo example for SDL with OpenGL

   Mike Wiering (mike@wieringsoftware.nl), July 2004
*/


#include <SDL/SDL.h>
#include <gl/gl.h>
#include <windows.h>
#include <mem.h>

#include "gfx\ScrollDemoGfx.c"  /* generated by Tile Studio */

#define TITLE "Tile Studio Scroll Demo"

#define SCREEN_WIDTH 640
#define SCREEN_HEIGHT 480

bool FullScreen = FALSE;

static int ScreenWidth  = SCREEN_WIDTH;
static int ScreenHeight = SCREEN_HEIGHT;

static int LayerCount = 0;

static int XView = 0;
static int YView = 0;
static int XScrollSpeed = 1;
static int YScrollSpeed = 1;

static int iMaxXScrollSpeed = 8;
static int iMaxYScrollSpeed = 8;
static int iActualXView = 0;
static int iActualYView = 0;

#define MAX_VALUE 0x7FFFFFFF

#define MAX_TILESET_COUNT 10
#define MAX_TEXTURE_COUNT 1000
#define MAX_TEX_W 64
#define MAX_TEX_H 64

GLuint Textures[MAX_TEXTURE_COUNT];
int TotalTextureCount = 0;

#define TEX_BUFFER_SIZE (MAX_TEX_W * MAX_TEX_H)
int TexBuffer[TEX_BUFFER_SIZE];



static inline int intdiv (int x, int y)
{
  return (x < 0)? (x - y + 1) / y : x / y;
}

static inline int Clip (int x, int min, int max)
{
  return (x < min)? min : ((x > max)? max : x);
}

static inline int Min (int x, int y)
{
  return (x < y)? x : y;
}

static inline int Max (int x, int y)
{
  return (x > y)? x : y;
}



/*
typedef struct GFX_TILESET_DATA
{
  int TileWidth;
  int TileHeight;
  int TileCount;
  int Sequences;
  int ** SequenceData;
  int * SequenceLength;
  int * SequencePosition;
  int * SequenceCounter;
  int * BitmapRGBAData;
} GFX_TILESET_DATA;
*/


typedef struct TILESET_DEF
{
  int iTexWidth;
  int iTexHeight;
  int iTileWidth;
  int iTileHeight;
  int iTileCount;
  GLuint *ipTextures;

  int iSequences;
  int *ipSequenceLength;
  int *ipSequencePosition;
  int *ipSequenceCounter;
  int **ippSequenceData;
  DWORD *dwpFillColor;
  bool bInitialized;
} TILESET_DEF;

static TILESET_DEF tdTileset[MAX_TILESET_COUNT];


int InitializeTileset (const struct GFX_TILESET_DATA *gfx)
{
  TILESET_DEF *tdp = &tdTileset[0];
  int w, h, txw, txh;
  int id = 0;
  int src, dst;
  GLuint glTex;
  
  while (tdp->bInitialized)
  {
    tdp++;
    id++;
    if (id >= MAX_TILESET_COUNT)
      return -1;
  }
  
  w = gfx->TileWidth;
  h = gfx->TileHeight;
  txw = MAX_TEX_W;
  txh = MAX_TEX_H;
  while (txw / 2 >= w)
    txw /= 2;
  while (txh / 2 >= h)
    txh /= 2;
   
  tdp->ipTextures = &Textures[TotalTextureCount];

  tdp->iTexWidth = txw;
  tdp->iTexHeight = txh;  
  tdp->iTileCount = gfx->TileCount;
  tdp->iTileWidth = w;
  tdp->iTileHeight = h;

  tdp->iSequences = gfx->Sequences;
  tdp->ippSequenceData = gfx->SequenceData;
  tdp->ipSequenceLength = gfx->SequenceLength;
  tdp->ipSequencePosition = gfx->SequencePosition;
  tdp->ipSequenceCounter = gfx->SequenceCounter;

  src = 0;
  for (int tile = 0; tile < gfx->TileCount; tile++)
  {
    memset (&TexBuffer, 0, TEX_BUFFER_SIZE);
    dst = 0;
    for (int j = 0; j < h; j++)
    {
      for (int i = 0; i < w; i++)
        TexBuffer[dst++] = gfx->BitmapRGBAData[src++];
      dst += txw - w;
    }

    glEnable (GL_TEXTURE_2D);
    glGenTextures (1, &glTex);
    glBindTexture (GL_TEXTURE_2D, glTex);

    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  
    glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, txw, txh, 0, GL_RGBA, GL_UNSIGNED_BYTE, &TexBuffer);

    
    tdp->ipTextures[tile] = glTex;
    if (++TotalTextureCount >= MAX_TEXTURE_COUNT)
      return -1;
  }
  
  tdp->bInitialized = 1;
  return id;
}


void FinalizeTileset (int id)
{
  TILESET_DEF *tdp = &tdTileset[id];
  if (tdp->bInitialized)
    for (int tile = 0; tile < tdp->iTileCount; tile++)
      glDeleteTextures (1, &tdp->ipTextures[tile]);
  tdp->bInitialized = 0;
}


void RunSequences ()
{
  for (int id = 0; id < MAX_TILESET_COUNT; id++)
  {
    TILESET_DEF *tdp = &tdTileset[id];
    if (tdp->bInitialized)
      for (int i = 0; i < tdp->iSequences; i++)
        if (--tdp->ipSequenceCounter[i] <= 0)
        {
          if (++tdp->ipSequencePosition[i] >= tdp->ipSequenceLength[i])
            tdp->ipSequencePosition[i] = 0;
          tdp->ipSequenceCounter[i] =
                tdp->ippSequenceData[i][2 * tdp->ipSequencePosition[i] + 1];
        }
  }
}


static unsigned short *sBoundMap = NULL;
static int BoundMapWidth = 0;
static int BoundMapHeight = 0;
static int BoundBlockWidth = 0;
static int BoundBlockHeight = 0;
static int TotalWidth = 0;
static int TotalHeight = 0;


static int InitBoundMap (unsigned short *map, int w, int h, int bid)
{
  TILESET_DEF *tdp = &tdTileset[bid];

  sBoundMap = map;
  BoundMapWidth = w;
  BoundMapHeight = h;
  BoundBlockWidth = tdp->iTileWidth;
  BoundBlockHeight = tdp->iTileHeight;
  TotalWidth = tdp->iTileWidth * w;
  TotalHeight = tdp->iTileHeight * h;

  XView = 0;
  YView = 0;

  return 0;
}


/* move XView and YView towards iActualXView and iActualYView */
int CorrectView ()
{
  if (iActualXView > XView)
    XView = Min (XView + iMaxXScrollSpeed, iActualXView);
  if (iActualYView > YView)
    YView = Min (YView + iMaxYScrollSpeed, iActualYView);
  if (iActualXView < XView)
    XView = Max (XView - iMaxXScrollSpeed, iActualXView);
  if (iActualYView < YView)
    YView = Max (YView - iMaxYScrollSpeed, iActualYView);

  if (XView > TotalWidth - ScreenWidth)
  {
    XView = TotalWidth - ScreenWidth;
    iActualXView = XView;
    XScrollSpeed = -XScrollSpeed;
  }
  if (YView > TotalHeight - ScreenHeight)
  {
    YView = TotalHeight - ScreenHeight;
    iActualYView = YView;
    YScrollSpeed = -YScrollSpeed;
  }

  if (XView < 0)
  {
    XView = 0;
    iActualXView = XView;
    XScrollSpeed = -XScrollSpeed;
  }
  if (YView < 0)
  {
    YView = 0;
    iActualYView = YView;
    YScrollSpeed = -YScrollSpeed;
  }
  return 0;
}

static int MoveScreenTo (int x, int y)
{
  XView = x;
  YView = y;
  iActualXView = XView;
  iActualYView = YView;

  CorrectView ();
  return 0;
}

#define MAX_LAYER 8

typedef struct LAYERMAPINFO
{
  int iTilesetID;
  int iMapWidth;
  int iMapHeight;
  signed short *sMap;
  int iScrollFunction;
  int iHShift;
  int iVShift;
} LAYERMAPINFO;

static LAYERMAPINFO lmiLayerMapInfo[MAX_LAYER];


static int InitLayerMap (int tilesetid,
                         int mapwidth, int mapheight,
                         signed short *map,
                         int scrollfunction,
                         int hshift,
                         int vshift)
{
  if (LayerCount < MAX_LAYER)
  {
    LAYERMAPINFO *lmip = &lmiLayerMapInfo[LayerCount];
    lmip->iTilesetID = tilesetid;
    lmip->sMap = map;
    lmip->iMapWidth = mapwidth;
    lmip->iMapHeight = mapheight;
    lmip->iScrollFunction = scrollfunction;
    lmip->iHShift = hshift;
    lmip->iVShift = vshift;
    LayerCount++;
    return 0;
  }
  else
    return -1;
}


int CalcLayerPos (int scrollfunction, int *x, int *y)
{
  switch (scrollfunction)
  {
    case 0:
    {
      *x = XView;
      *y = YView;
    }
    break;
    default:
    {
      *x = XView / scrollfunction;
      *y = YView / scrollfunction;
    }
  }
  return 0;
}


void DrawTile (GLuint tex, int x, int y, int w, int h, float x1, float y1, float x2, float y2, bool mir, bool ups)
{
  glEnable (GL_TEXTURE_2D);
  glBindTexture (GL_TEXTURE_2D, tex);
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  if (mir) { float tmp = x1;  x1 = x2;  x2 = tmp; }    
  if (ups) { float tmp = y1;  y1 = y2;  y2 = tmp; }    

  glBegin (GL_QUADS);
    glTexCoord2f (x1, y1);  glVertex2f (x, y);
    glTexCoord2f (x2, y1);  glVertex2f (x + w, y);
    glTexCoord2f (x2, y2);  glVertex2f (x + w, y + h);
    glTexCoord2f (x1, y2);  glVertex2f (x, y + h);
  glEnd ();
}


int DrawLayer (int layer)
{
  int xstart, ystart;
  int x, y, w, h;
  int mapx, mapy;
  int mapw, maph;
  int mappos, tile;
  int fx;
  int maxtile;
  BOOL mir, ups;
  LAYERMAPINFO *lmip = &lmiLayerMapInfo[layer];
/*
  int front = 0;
  int frontblocks = 0;
*/

  if ((layer < LayerCount) && (lmip))
  {
    TILESET_DEF *tdp = &tdTileset[lmip->iTilesetID];
    
    if (tdp)
    {
      w = tdp->iTileWidth;
      h = tdp->iTileHeight;
      mapw = lmip->iMapWidth;
      maph = lmip->iMapHeight;
      maxtile = tdp->iTileCount;

      // calculate layer position
      CalcLayerPos (lmip->iScrollFunction, &xstart, &ystart);
      xstart += lmip->iHShift;
      ystart += lmip->iVShift;
      while (xstart < 0)
        xstart += mapw * w;
      while (ystart < 0)
        ystart += maph * h;

      mapy = ystart / h;
      for (y = -(ystart % h); y < ScreenHeight; )
      {
        mapx = xstart / w;
        for (x = -(xstart % w); x < ScreenWidth; )
        {
          // calculate position in map, read tile number
          mappos = ((mapx % mapw) + ((mapy % maph) * mapw));
          tile = *(signed short *) &lmip->sMap[mappos];

          if (tile < 0)  // sequence ID
          {
            int seq = -tile - 1;
            if (seq < tdp->iSequences)
            {
              int pos = tdp->ipSequencePosition[seq];
              int *seqdata = tdp->ippSequenceData[seq];
              tile = seqdata[2 * pos];
            }
            else
            {
              seq -= tdp->iSequences;
              tile = 0;
            }
          }

        /*
          if (front)
          {
            if (tile >= 0x4000)
              tile -= (0x4000 - 1);
            else
              tile = 0;
          }
          else
            if (tile >= 0x4000)
            {
              tile -= 0x4000;
              frontblocks++;
            }
        */

          tile--;
          fx = (tile / maxtile) & 3;

          if (tile >= 0)
          {
            mir = FALSE;
            ups = FALSE;
            if ((fx & 1) == 1)
              mir = TRUE;
            if ((fx & 2) == 2)
              ups = TRUE;

            DrawTile(tdp->ipTextures[tile % maxtile], x, y, w, h, 0, 0,
                       (float) tdp->iTileWidth / tdp->iTexWidth, (float) tdp->iTileHeight / tdp->iTexHeight, 
                         mir, ups);
          }
          x += w;
          mapx++;
        }

        y += h;
        mapy++;
      }
    }
  }
  
/*
  if (frontblocks > 0)
    return 1;
  else
*/
    return 0;
}


void ClearScreen ()
{
  glClearColor (0.0, 0.0, 0.0, 1);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


void SetViewPort (int x, int y, int w, int h)
{
  glViewport (x, ScreenHeight - y - h, w, h);
  glLoadIdentity ();
  glOrtho (0, w, h, 0, -1, 1);
}


void RunFrame ()
{
  static int initialized = 0;
  int x = XView / BoundBlockWidth;
  int y = YView / BoundBlockHeight;
  int w = ScreenWidth / BoundBlockWidth;
  int h = ScreenHeight / BoundBlockHeight;

  if (!initialized)
  {
   // XScrollSpeed = 0;
   // YScrollSpeed = 0;
    CorrectView ();

    initialized++;
  }

  DrawLayer (0);
  DrawLayer (1);

  RunSequences ();

  iActualXView += XScrollSpeed;
  iActualYView += YScrollSpeed;
  CorrectView ();
}



void InitializeTilesets ()
{
  int id = 0;
  memset (&tdTileset, 0, sizeof (tdTileset));
  LayerCount = 0;

  id = InitializeTileset (&GfxBackGrTilesetData);
  InitLayerMap (id, GfxBackGrMap1MapWidth, GfxBackGrMap1MapHeight,
                &BackGrMap1MapData[0][0], 3, 0, 0);

  id = InitializeTileset (&GfxBlocksTilesetData);
  InitLayerMap (id, GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight,
                &BlocksLevel1MapData[0][0], 1, 0, 0);

  InitBoundMap (&BlocksLevel1BoundMapData[0][0],
                GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight, id);
}



int init ()
{
  int flags = SDL_OPENGL | SDL_HWSURFACE;
  if (FullScreen)
    flags |= SDL_NOFRAME | SDL_FULLSCREEN;
  
  SDL_SetVideoMode (ScreenWidth, ScreenHeight, 0, flags);

  SetViewPort (0, 0, ScreenWidth, ScreenHeight);
  
  ClearScreen ();
  SDL_GL_SwapBuffers ();
  ClearScreen ();
  SDL_GL_SwapBuffers ();

  glDisable (GL_DEPTH_TEST);
  glDisable (GL_CULL_FACE);
  glEnable (GL_TEXTURE_2D);

  InitializeTilesets ();
}


int finalize ()
{
  for (int i = 0; i < MAX_TILESET_COUNT; i++)
    FinalizeTileset (i);
}



int main (int Arg_N, char ** Arg_V)
{
  SDL_Event event;
  float theta = 0.0f;
  
  int count = 50;
  float theta_tmp;

  SDL_Init (SDL_INIT_VIDEO);

  SDL_WM_SetCaption (TITLE, 0);

  init ();


  for (int done = 0; !done;)
  {
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    RunFrame ();
    
    SDL_GL_SwapBuffers ();
    SDL_PollEvent (&event);
    switch (event.type)
    {
      case SDL_KEYDOWN:
        switch (event.key.keysym.sym)
        {
          case SDLK_ESCAPE:
            done = TRUE;
            break;
          case SDLK_LEFT:
            XScrollSpeed--;
            break;
          case SDLK_RIGHT:
            XScrollSpeed++;
            break;
          case SDLK_UP:
            YScrollSpeed--;
            break;
          case SDLK_DOWN:
            YScrollSpeed++;
            break;
        }
    }  
    event.key.keysym.sym = SDLK_SPACE;  // erase last key
  }

  finalize ();

  SDL_Quit ();
  return (0);
}
